/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.dns.protocol;

import java.io.IOException;
import java.net.InetAddress;

import org.apache.log4j.Logger;

import br.registro.dnsshim.common.server.AbstractDecoderOutput;
import br.registro.dnsshim.common.server.DnsshimProtocolException;
import br.registro.dnsshim.common.server.IoSession;
import br.registro.dnsshim.common.server.ProtocolStatusCode;
import br.registro.dnsshim.common.server.Server.TransportType;
import br.registro.dnsshim.domain.Opt;
import br.registro.dnsshim.domain.ResourceRecord;
import br.registro.dnsshim.domain.RrType;
import br.registro.dnsshim.domain.Soa;
import br.registro.dnsshim.util.ByteUtil;
import br.registro.dnsshim.util.DomainNameUtil;
import br.registro.dnsshim.util.ResourceRecordUtil;
import br.registro.dnsshim.xfrd.domain.ZoneData;
import br.registro.dnsshim.xfrd.domain.ZoneInfo;
import br.registro.dnsshim.xfrd.domain.logic.ZoneDataCache;
import br.registro.dnsshim.xfrd.domain.logic.ZoneInfoCache;

public class DecoderOutput extends AbstractDecoderOutput {

	private static final Logger logger = Logger.getLogger(DecoderOutput.class);

	@Override
	public void doOutput(Object message, IoSession session)
	throws DnsshimProtocolException {
		if (!(message instanceof DnsMessage)) {
			throw new DnsshimProtocolException(ProtocolStatusCode.DNS_MESSAGE_FORMERR, "Invalid DNS message");
		}

		DnsMessage request = (DnsMessage) message;
		DnsPacket dnsPacket = request.getFirstPacket();
		DnsQuestion question = dnsPacket.getQuestion();
		if (logger.isDebugEnabled()) {
			logger.debug("Got query: " + question);
		}

		// Add the same OPT record of the question in the answer
		Opt opt = dnsPacket.getOpt();

		try {
			String qname = DomainNameUtil.trimDot(question.getQname());
			String zonename = ResourceRecordUtil.findZonename(qname);
			if (zonename.isEmpty()) {
				dnsPacket.getHeader().setResponseCode(ResponseCode.REFUSED);
				throw new DnsshimProtocolException(ProtocolStatusCode.NOT_AUTHORITATIVE, "Not authoritative for " + qname,
						dnsPacket);
			}


			
			ZoneInfoCache zoneInfoCache = ZoneInfoCache.getInstance();
			ZoneInfo zoneInfo = zoneInfoCache.get(zonename);
			if (zoneInfo == null) {
				dnsPacket.getHeader().setResponseCode(ResponseCode.REFUSED);
				throw new DnsshimProtocolException(ProtocolStatusCode.NOT_AUTHORITATIVE, "Not authoritative for " + qname,
						dnsPacket);
			}
			
			InetAddress remote = session.getRemoteAddress().getAddress();
			if ((question.getQtype() == QueryType.AXFR ||
					question.getQtype() == QueryType.IXFR) &&
					!remote.getCanonicalHostName().equals("localhost") && 
					!zoneInfo.hasSlave(remote.getHostAddress())) {
				dnsPacket.getHeader().setResponseCode(ResponseCode.REFUSED);
				throw new DnsshimProtocolException(ProtocolStatusCode.NOT_AUTHORITATIVE, "Unauthorized request from " +
						remote.getHostAddress(), dnsPacket);
			}

			DnsMessageBuilder messageBuilder = null;

			ZoneDataCache zoneDataCache = ZoneDataCache.getInstance();
			ZoneData zoneData = zoneDataCache.get(zonename);
			if (zoneData == null) {
				dnsPacket.getHeader().setResponseCode(ResponseCode.REFUSED);
				throw new DnsshimProtocolException(ProtocolStatusCode.REFUSED_OPERATION, "Error loading zone " + zonename +
						". Maybe it hasn't been published yet.", dnsPacket);
			}

			switch (question.getQtype()) {
			case AXFR:
				if (logger.isInfoEnabled()) {
					String remoteAddr = session.getRemoteAddress().getAddress().getHostAddress();
					logger.info("Got AXFR request for zone " + zonename + " from " + remoteAddr);
				}
				messageBuilder = new AxfrMessageBuilder();
				break;
			case IXFR:
				if (logger.isInfoEnabled()) {
					String remoteAddr = session.getRemoteAddress().getAddress().getHostAddress();
					logger.info("Got IXFR request for zone " + zonename + " from " + remoteAddr);
				}
				int serial = 0;
				if (!dnsPacket.getAuthority().isEmpty()) {
					for (ResourceRecord rr : dnsPacket.getAuthority()) {
						if (rr.getType() == RrType.SOA) {
							Soa soa = (Soa) rr;
							serial = soa.getSerial();
							break;
						}
					}
				}

				if (zoneData.containsIxfrSerial(serial)) {
					messageBuilder = new IxfrMessageBuilder(serial, opt);
				} else {
					messageBuilder = new AxfrMessageBuilder();
				}

				break;
			case ANY:
				messageBuilder = new AnyMessageBuilder(opt);
				break;
			default:
				messageBuilder = new RecordMessageBuilder(opt);
				break;
			}

			if (session.getTransportType() == TransportType.TCP) {
				messageBuilder.setMaxPacketSize(DnsPacket.MAX_TCP_PACKET_SIZE);
			} else {
				int udpSize = DnsPacket.MAX_UDP_PACKET_SIZE;
				if (dnsPacket.getOpt() != null) {
					udpSize = ByteUtil.toUnsigned(dnsPacket.getOpt().getUdpPayloadSize());
				}

				messageBuilder.setMaxPacketSize(udpSize);
			}

			messageBuilder.setZoneData(zoneData);
			messageBuilder.setRequestPacket(dnsPacket);
			messageBuilder.process();
			
			session.setAttachment(request.getFirstPacket().getTsig());
			DnsMessage response = messageBuilder.getDnsMessage();
			getEncoder().encode(response, session);
			
		} catch (IOException e) {
			dnsPacket.getHeader().setResponseCode(ResponseCode.SERVFAIL);
			throw new DnsshimProtocolException(ProtocolStatusCode.DNS_SERVER_FAILURE, e.getMessage(), request);
		}	
	}
}