/* Copyright (C) 2009 Registro.br. All rights reserved. 
* 
* Redistribution and use in source and binary forms, with or without 
* modification, are permitted provided that the following conditions are 
* met:
* 1. Redistribution of source code must retain the above copyright 
*    notice, this list of conditions and the following disclaimer.
* 2. Redistributions in binary form must reproduce the above copyright
*    notice, this list of conditions and the following disclaimer in the
*    documentation and/or other materials provided with the distribution.
* 
* THIS SOFTWARE IS PROVIDED BY REGISTRO.BR ``AS IS'' AND ANY EXPRESS OR
* IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
* WARRANTIE OF FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO
* EVENT SHALL REGISTRO.BR BE LIABLE FOR ANY DIRECT, INDIRECT,
* INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
* BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
* OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
* ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
* USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
* DAMAGE.
 */
package br.registro.dnsshim.xfrd.domain;

import java.util.ArrayList;
import java.util.List;

import br.registro.dnsshim.domain.logic.Cacheable;
import br.registro.dnsshim.util.ByteUtil;
import br.registro.dnsshim.xfrd.domain.logic.SlaveGroupCache;

/*
 * All numeric time values here are expressed in seconds.
 * Care must be taken, for the Java API stores those values in milliseconds.
 */
public class ZoneInfo implements Cacheable<String>{	
	public static final String DEFAULT_ZONENAME = "";
	public static final int DEFAULT_TTL_ZONE = 86400;
	public static final int DEFAULT_TTL_DNSKEY = 86400;
	public static final int DEFAULT_SOA_MINIMUM = 900;
	public static final int DEFAULT_EXPIRATION = 30 * 24 * 60 * 60; // 1 month
	public static final int MIN_EXPIRATION = 2 * 86400; // 2 days
	public static final int MAX_EXPIRATION = 5 * 31536000; // 5 years
	public static final int DEFAULT_MAX_INCREMENTS = 5;
	
	public static final boolean DEFAULT_SIGNED = true;
	public static final boolean DEFAULT_VERIFY_SIGNATURE = false;
	
	private String zonename = DEFAULT_ZONENAME;
	private int ttlZone = DEFAULT_TTL_ZONE;
	private int ttlDnskey = DEFAULT_TTL_DNSKEY;
	private int soaMinimum = DEFAULT_SOA_MINIMUM;
	private int expiration = DEFAULT_EXPIRATION;
	
	private int maxIncrements = DEFAULT_MAX_INCREMENTS; 
	
	private boolean signed = DEFAULT_SIGNED;
	private boolean verifySignature = DEFAULT_VERIFY_SIGNATURE;
	
	private long nextPublication;
	
	private List<String> noneKeys = new ArrayList<String>();
	private List<String> publishKeys = new ArrayList<String>();
	private List<String> signKeys = new ArrayList<String>();

	/*
	 * Old keys, that must be kept in order to keep track of changes to the keys
	 */
	private List<String> oldNoneKeys = new ArrayList<String>();
	private List<String> oldPublishKeys = new ArrayList<String>();
	private List<String> oldSignKeys = new ArrayList<String>();
	
	private List<String> slaveGroups = new ArrayList<String>();
	private List<String> users = new ArrayList<String>();

	public String getZonename() {
		return zonename;
	}

	public void setZonename(String zonename) {
		this.zonename = zonename;
	}

	public int getTtlZone() {
		return ttlZone;
	}

	public void setTtlZone(int ttlZone) {
		this.ttlZone = ttlZone;
	}

	public int getTtlDnskey() {
		return ttlDnskey;
	}

	public void setTtlDnskey(int ttlDnskey) {
		this.ttlDnskey = ttlDnskey;
	}

	public int getSoaMinimum() {
		return soaMinimum;
	}

	public void setSoaMinimum(int soaMinimum) {
		this.soaMinimum = soaMinimum;
	}

	public long getNextPublication() {
		return nextPublication;
	}

	public void setNextPublication(long nextPublication) {
		this.nextPublication = nextPublication;
	}

	public int getExpiration() {
		return expiration;
	}

	public void setExpiration(int expiration) {
		if (expiration < 0 || expiration > MAX_EXPIRATION) {
			this.expiration = MAX_EXPIRATION;
			return;
		}
		
		if (expiration < MIN_EXPIRATION) {
			this.expiration = MIN_EXPIRATION;
			return;
		}

		this.expiration = expiration;
	}

	public void setMaxIncrements(int maxIncrements) {
		if (maxIncrements < 0) {
			throw new IllegalArgumentException("Invalid value for MaxIncrements!");
		}
		
		this.maxIncrements = maxIncrements;
	}

	public int getMaxIncrements() {
		return maxIncrements;
	}
	
	public boolean isSigned() {
		return signed;
	}

	public void setSigned(boolean signed) {
		this.signed = signed;
	}

	public boolean isVerifySignature() {
		return verifySignature;
	}

	public void setVerifySignature(boolean verifySignature) {
		this.verifySignature = verifySignature;
	}
	
	public List<String> getKeys(DnskeyStatus status) {
		switch (status) {
		case NONE:
			return noneKeys;
		case PUBLISH:
			return publishKeys;
		case SIGN:
			return signKeys;
		}
		
		return new ArrayList<String>();
	}
	
	public List<String> getAllKeys() {
		List<String> keys = new ArrayList<String>();
		keys.addAll(noneKeys);
		keys.addAll(publishKeys);
		keys.addAll(signKeys);
		return keys;
	}
	
	public String getSignKey(short keyTag) {
		String keyTagStr = "" + ByteUtil.toUnsigned(keyTag);
		for (String key : signKeys) {
			if (key.contains(keyTagStr)) {
				return key;
			}
		}
		
		return null;
	}
	
	public void addOldKey(String key, DnskeyStatus status) {
		switch(status) {
		case NONE:
			if (!oldNoneKeys.contains(key)) {
				oldNoneKeys.add(key);
			}
			break;
		case PUBLISH:
			if (!oldPublishKeys.contains(key)) {
				oldPublishKeys.add(key);
			}
			break;
		case SIGN:
			if (!oldSignKeys.contains(key)) {
				oldSignKeys.add(key);
			}
			break;
		}
	}
	
	public List<String> getOldKeys(DnskeyStatus status) {
		switch (status) {
		case NONE:
			return oldNoneKeys;
		case PUBLISH:
			return oldPublishKeys;
		case SIGN:
			return oldSignKeys;
		}
		
		return new ArrayList<String>();
	}
	
	public boolean hasSlave(String address) {
		SlaveGroupCache slaveGroupCache = SlaveGroupCache.getInstance();
		for(String groupName : slaveGroups) {
			SlaveGroup group = slaveGroupCache.get(groupName);
			if (group != null && group.hasSlave(address)) {
				return true;
			}
		}
		
		return false;
	}
	
	public List<String> getSlaveGroups() {
		return slaveGroups;
	}

	public void assignSlaveGroup(String group) {
		this.slaveGroups.add(group);
	}
	
	public boolean unassignSlaveGroup(String slave) {
		return slaveGroups.remove(slave);
	}
	
	public List<String> getUsers() {
		return users;
	}

	public void setUsers(List<String> users) {
		this.users = users;
	}

	public void addUser(String username) {
		this.users.add(username);
	}
	
	public void updateKeys() {
		oldNoneKeys.clear();
		oldPublishKeys.clear();
		oldSignKeys.clear();
		
		oldNoneKeys.addAll(noneKeys);
		oldPublishKeys.addAll(publishKeys);
		oldSignKeys.addAll(signKeys);
	}
	
	public boolean isModifiedKeys() {
		// Keys with status NONE are not taken into account here
		return !(publishKeys.equals(oldPublishKeys) &&
				signKeys.equals(oldSignKeys));
	}

	public DnskeyStatus hasKey(String key) {
		if (noneKeys.contains(key)) {
			return DnskeyStatus.NONE;
		} else if (publishKeys.contains(key)) {
			return DnskeyStatus.PUBLISH;
		} else if (signKeys.contains(key)) {
			return DnskeyStatus.SIGN;
		}
		
		return null;
	}
	
	public boolean addKey(String key, DnskeyStatus status) {
		if (hasKey(key) != null) {
			return false;
		}
		
		switch (status) {
		case NONE:
			noneKeys.add(key);
			break;
		case PUBLISH:
			publishKeys.add(key);
			break;
		case SIGN:
			signKeys.add(key);
			break;
		}
		return true;
	}
	
	public boolean removeKey(String key, DnskeyStatus status) {
		switch (status) {
		case NONE:
			return noneKeys.remove(key);
		case PUBLISH:
			return publishKeys.remove(key);
		case SIGN:
			return signKeys.remove(key);
		}
		
		return false;
	}

	@Override
	public String getCacheKey() {
		return this.zonename;
	}
	
	@Override
	public boolean equals(Object obj) {
		if (obj instanceof ZoneInfo) {
			ZoneInfo z = (ZoneInfo) obj;
			
			if (z.zonename.equals(zonename) && 
					z.ttlZone == ttlZone && 
					z.ttlDnskey == ttlDnskey && 
					z.soaMinimum == soaMinimum && 
					z.expiration == expiration && 
					
					z.maxIncrements == maxIncrements &&
					
					z.signed == signed &&
					z.verifySignature == verifySignature &&
					
					z.noneKeys.equals(noneKeys) &&
					z.publishKeys.equals(publishKeys) &&
					z.signKeys.equals(signKeys) &&
					
					z.oldNoneKeys.equals(oldNoneKeys) &&
					z.oldPublishKeys.equals(oldPublishKeys) &&
					z.oldSignKeys.equals(oldSignKeys) &&
					
					z.slaveGroups.equals(slaveGroups) && 
					z.users.equals(users)) {
				return true;
			}
		}
		
		return false;
	}
}
